@@ -60,12 +60,15 @@
 #include        "string.h"              /* strings                         */
 
 #include        "code.h"                /* coder                           */
-#include        "vars.h"                /* variables                       */
 
 #include        "stats.h"               /* statements                      */
 
 #include        "saveload.h"            /* saving and loading              */
+#include        "tls.h"                 /* thread-local storage            */
 
+#include        "vars.h"                /* variables                       */
+
+#include <assert.h>
 
 /****************************************************************************
 **
@@ -274,19 +277,38 @@ Obj DoWrap6args (
 
 *F  DoFail0args( <self> )  . . . . . .  fail a function call with 0 arguments
 **
-**  'DoWrap<i>args' accepts the <i> arguments <arg1>, <arg2>,  and so on, and
+**  'DoFail<i>args' accepts the <i> arguments <arg1>, <arg2>,  and so on, and
 **  signals an error,  because  the  function for  which  they  are installed
 **  expects another number of arguments.  'DoFail<i>args' are the handlers in
 **  the other slots of a function.
 */
+
+/* Pull this out to avoid repetition, since it gets a little more complex in 
+   the presence of partially variadic functions */
+
+Obj NargError( Obj func, Int actual) {
+  Int narg = NARG_FUNC(func);
+
+  if (narg >= 0) {
+    assert(narg != actual);
+    return ErrorReturnObj(
+			  "Function: number of arguments must be %d (not %d)",
+			  narg, actual,
+			  "you can replace the argument list <args> via 'return <args>;'" );
+  } else {
+    assert(-narg-1 > actual);
+    return ErrorReturnObj(
+        "Function: number of arguments must be at least %d (not %d)",
+        -narg-1, actual,
+        "you can replace the argument list <args> via 'return <args>;'" );
+  }
+}
+
 Obj DoFail0args (
     Obj                 self )
 {
     Obj                 argx;           /* arguments list (to continue)    */
-    argx = ErrorReturnObj(
-        "Function: number of arguments must be %d (not %d)",
-        (Int)NARG_FUNC( self ), 0L,
-        "you can replace the argument list <args> via 'return <args>;'" );
+    argx =NargError(self, 0);
     return FuncCALL_FUNC_LIST( (Obj)0, self, argx );
 }
 
@@ -300,10 +322,7 @@ Obj DoFail1args (
     Obj                 arg1 )
 {
     Obj                 argx;           /* arguments list (to continue)    */
-    argx = ErrorReturnObj(
-        "Function: number of arguments must be %d (not %d)",
-        (Int)NARG_FUNC( self ), 1L,
-        "you can replace the argument list <args> via 'return <args>;'" );
+    argx =NargError(self, 1);
     return FuncCALL_FUNC_LIST( (Obj)0, self, argx );
 }
 
@@ -318,10 +337,7 @@ Obj DoFail2args (
     Obj                 arg2 )
 {
     Obj                 argx;           /* arguments list (to continue)    */
-    argx = ErrorReturnObj(
-        "Function: number of arguments must be %d (not %d)",
-        (Int)NARG_FUNC( self ), 2L,
-        "you can replace the argument list <args> via 'return <args>;'" );
+    argx =NargError(self, 2);
     return FuncCALL_FUNC_LIST( (Obj)0, self, argx );
 }
 
@@ -337,10 +353,7 @@ Obj DoFail3args (
     Obj                 arg3 )
 {
     Obj                 argx;           /* arguments list (to continue)    */
-    argx = ErrorReturnObj(
-        "Function: number of arguments must be %d (not %d)",
-        (Int)NARG_FUNC( self ), 3L,
-        "you can replace the argument list <args> via 'return <args>;'" );
+    argx =NargError(self, 3);
     return FuncCALL_FUNC_LIST( (Obj)0, self, argx );
 }
 
@@ -357,10 +370,7 @@ Obj DoFail4args (
     Obj                 arg4 )
 {
     Obj                 argx;           /* arguments list (to continue)    */
-    argx = ErrorReturnObj(
-        "Function: number of arguments must be %d (not %d)",
-        (Int)NARG_FUNC( self ), 4L,
-        "you can replace the argument list <args> via 'return <args>;'" );
+    argx =NargError(self, 4);
     return FuncCALL_FUNC_LIST( (Obj)0, self, argx );
 }
 
@@ -378,10 +388,7 @@ Obj DoFail5args (
     Obj                 arg5 )
 {
     Obj                 argx;           /* arguments list (to continue)    */
-    argx = ErrorReturnObj(
-        "Function: number of arguments must be %d (not %d)",
-        (Int)NARG_FUNC( self ), 5L,
-        "you can replace the argument list <args> via 'return <args>;'" );
+    argx =NargError(self, 5);
     return FuncCALL_FUNC_LIST( (Obj)0, self, argx );
 }
 
@@ -400,10 +407,7 @@ Obj DoFail6args (
     Obj                 arg6 )
 {
     Obj                 argx;           /* arguments list (to continue)    */
-    argx = ErrorReturnObj(
-        "Function: number of arguments must be %d (not %d)",
-        (Int)NARG_FUNC( self ), 6L,
-        "you can replace the argument list <args> via 'return <args>;'" );
+    argx =NargError(self, 6);
     return FuncCALL_FUNC_LIST( (Obj)0, self, argx );
 }
 
@@ -417,10 +421,7 @@ Obj DoFailXargs (
     Obj                 args )
 {
     Obj                 argx;           /* arguments list (to continue)    */
-    argx = ErrorReturnObj(
-        "Function: number of arguments must be %d (not %d)",
-        (Int)NARG_FUNC( self ), LEN_LIST( args ),
-        "you can replace the argument list <args> via 'return <args>;'" );
+    argx =NargError(self, LEN_LIST(args));
     return FuncCALL_FUNC_LIST( (Obj)0, self, argx );
 }
 
@@ -1140,7 +1141,7 @@ Obj NewFunctionT (
     func = NewBag( type, size );
 
     /* create a function with a fixed number of arguments                  */
-    if ( narg != -1 ) {
+    if ( narg >= 0 ) {
         HDLR_FUNC(func,0) = DoFail0args;
         HDLR_FUNC(func,1) = DoFail1args;
         HDLR_FUNC(func,2) = DoFail2args;
@@ -1154,20 +1155,21 @@ Obj NewFunctionT (
 
     /* create a function with a variable number of arguments               */
     else {
-        HDLR_FUNC(func,0) = DoWrap0args;
-        HDLR_FUNC(func,1) = DoWrap1args;
-        HDLR_FUNC(func,2) = DoWrap2args;
-        HDLR_FUNC(func,3) = DoWrap3args;
-        HDLR_FUNC(func,4) = DoWrap4args;
-        HDLR_FUNC(func,5) = DoWrap5args;
-        HDLR_FUNC(func,6) = DoWrap6args;
-        HDLR_FUNC(func,7) = hdlr;
+      HDLR_FUNC(func,0) = (narg >= -1) ? DoWrap0args : DoFail0args;
+      HDLR_FUNC(func,1) = (narg >= -2) ? DoWrap1args : DoFail1args;
+      HDLR_FUNC(func,2) = (narg >= -3) ? DoWrap2args : DoFail2args;
+      HDLR_FUNC(func,3) = (narg >= -4) ? DoWrap3args : DoFail3args;
+      HDLR_FUNC(func,4) = (narg >= -5) ? DoWrap4args : DoFail4args;
+      HDLR_FUNC(func,5) = (narg >= -6) ? DoWrap5args : DoFail5args;
+      HDLR_FUNC(func,6) = (narg >= -7) ? DoWrap6args : DoFail6args;
+      HDLR_FUNC(func,7) = hdlr;
     }
 
-    /* enter the the arguments and the names                               */
+    /* enter the arguments and the names                               */
     NAME_FUNC(func) = name;
     NARG_FUNC(func) = narg;
     NAMS_FUNC(func) = nams;
+    if (nams) MakeBagPublic(nams);
     CHANGED_BAG(func);
 
     /* enter the profiling bag                                             */
@@ -1203,10 +1205,29 @@ Obj NewFunctionCT (
     ObjFunc             hdlr )
 {
     Obj                 name_o;         /* name as an object               */
-    Obj                 nams_o;         /* nams as an object               */
-    UInt                 len;            /* length                          */
-    UInt                 i, k, l;        /* loop variables                  */
 
+    /* convert the name to an object                                       */
+    C_NEW_STRING_DYN(name_o, name_c);
+    RetypeBag(name_o, T_STRING+IMMUTABLE);
+
+    /* make the function                                                   */
+    return NewFunctionT( type, size, name_o, narg, ArgStringToList( nams_c ), hdlr );
+}
+    
+
+/****************************************************************************
+**
+*F  ArgStringToList( <nams_c> )
+**
+** 'ArgStringToList' takes a C string <nams_c> containing a list of comma
+** separated argument names, and turns it into a plist of strings, ready
+** to be passed to 'NewFunction' as <nams>.
+*/
+Obj ArgStringToList(const Char *nams_c) {
+    Obj                 tmp;            /* argument name as an object      */
+    Obj                 nams_o;         /* nams as an object               */
+    UInt                len;            /* length                          */
+    UInt                i, k, l;        /* loop variables                  */
 
     /* convert the arguments list to an object                             */
     len = 0;
@@ -1224,23 +1245,18 @@ Obj NewFunctionCT (
             k++;
         }
         l = k;
-        while ( nams_c[l]!=' ' && nams_c[l]!=',' && nams_c[l]!='\0' ) {
+        while ( nams_c[l] != ' ' && nams_c[l] != ',' && nams_c[l] != '\0' ) {
             l++;
         }
-        C_NEW_STRING(name_o, l-k, nams_c+k);
-        RESET_FILT_LIST( name_o, FN_IS_MUTABLE );
-        SET_ELM_PLIST( nams_o, i, name_o );
+        C_NEW_STRING( tmp, l - k, nams_c + k );
+        RetypeBag( tmp, T_STRING+IMMUTABLE );
+        SET_ELM_PLIST( nams_o, i, tmp );
         k = l;
     }
 
-    /* convert the name to an object                                       */
-    C_NEW_STRING_DYN(name_o, name_c);
-    RESET_FILT_LIST( name_o, FN_IS_MUTABLE );
-
-    /* make the function                                                   */
-    return NewFunctionT( type, size, name_o, narg, nams_o, hdlr );
+    return nams_o;
 }
-    
+
 
 /****************************************************************************
 **
@@ -1251,9 +1267,9 @@ Obj NewFunctionCT (
 /****************************************************************************
 **
 
-*F  TypeFunction( <func> )  . . . . . . . . . . . . . . .  kind of a function
+*F  TypeFunction( <func> )  . . . . . . . . . . . . . . .  type of a function
 **
-**  'TypeFunction' returns the kind of the function <func>.
+**  'TypeFunction' returns the type of the function <func>.
 **
 **  'TypeFunction' is the function in 'TypeObjFuncs' for functions.
 */
@@ -1283,7 +1299,9 @@ void PrintFunction (
     Int                 nloc;           /* number of locals                */
     Obj                 oldLVars;       /* terrible hack                   */
     UInt                i;              /* loop variable                   */
+    UInt                isvarg;         /* does function have varargs?     */
 
+    isvarg = 0;
 
     if ( IS_OPERATION(func) ) {
       CALL_1ARGS( PrintOperation, func );
@@ -1294,12 +1312,20 @@ void PrintFunction (
     Pr("%5>function%< ( %>",0L,0L);
 
     /* print the arguments                                                 */
-    narg = (NARG_FUNC(func) == -1 ? 1 : NARG_FUNC(func));
+    narg = NARG_FUNC(func);
+    if (narg < 0) {
+      isvarg = 1;
+      narg = -narg;
+    }
+    
     for ( i = 1; i <= narg; i++ ) {
         if ( NAMS_FUNC(func) != 0 )
             Pr( "%I", (Int)NAMI_FUNC( func, (Int)i ), 0L );
         else
             Pr( "<<arg-%d>>", (Int)i, 0L );
+        if(isvarg && i == narg) {
+            Pr("...", 0L, 0L);
+        }
         if ( i != narg )  Pr("%<, %>",0L,0L);
     }
     Pr(" %<)",0L,0L);
@@ -1325,7 +1351,7 @@ void PrintFunction (
             Pr("<<kernel or compiled code>>",0L,0L);
         }
         else {
-            SWITCH_TO_NEW_LVARS( func, NARG_FUNC(func), NLOC_FUNC(func),
+            SWITCH_TO_NEW_LVARS( func, narg, NLOC_FUNC(func),
                                  oldLVars );
             PrintStat( FIRST_STAT_CURR_FUNC );
             SWITCH_TO_OLD_LVARS( oldLVars );
@@ -1555,7 +1581,7 @@ Obj FuncNAME_FUNC (
         name = NAME_FUNC(func);
         if ( name == 0 ) {
             C_NEW_STRING_CONST(name, "unknown");
-            RESET_FILT_LIST( name, FN_IS_MUTABLE );
+            RetypeBag(name, T_STRING+IMMUTABLE);
             NAME_FUNC(func) = name;
             CHANGED_BAG(func);
         }
@@ -1571,9 +1597,10 @@ Obj FuncSET_NAME_FUNC(
                       Obj func,
                       Obj name )
 {
-  while (!IsStringConv(name))
+  while (!IsStringConv(name)) {
     name = ErrorReturnObj("SET_NAME_FUNC( <func>, <name> ): <name> must be a string, not a %s",
                           (Int)TNAM_OBJ(name), 0, "YOu can return a new name to continue");
+  }
   if (TNUM_OBJ(func) == T_FUNCTION ) {
     NAME_FUNC(func) = name;
     CHANGED_BAG(func);
@@ -1768,8 +1795,11 @@ Obj FuncFILENAME_FUNC(Obj self, Obj func) {
 
     if (BODY_FUNC(func)) {
         Obj fn =  FILENAME_BODY(BODY_FUNC(func));
-        if (fn)
+#ifndef WARD_ENABLED
+        if (fn) {
             return fn;
+        }
+#endif
     }
     return Fail;
 }
@@ -2017,7 +2047,7 @@ static Int InitKernel (
     InfoBags[ T_FUNCTION ].name = "function";
     InitMarkFuncBags( T_FUNCTION , MarkAllSubBags );
 
-    /* install the kind function                                           */
+    /* install the type functions                                          */
     ImportGVarFromLibrary( "TYPE_FUNCTION",  &TYPE_FUNCTION  );
     ImportGVarFromLibrary( "TYPE_OPERATION", &TYPE_OPERATION );
     TypeObjFuncs[ T_FUNCTION ] = TypeFunction;
@@ -2106,7 +2136,6 @@ static StructInitInfo module = {
 
 StructInitInfo * InitInfoCalls ( void )
 {
-    FillInVersion( &module );
     return &module;
 }
 
