@@ -36,6 +36,8 @@
 
 #include        "gap.h"                 /* error handling, initialisation  */
 
+#include        "code.h"                /* coder                           */
+
 #include        "gvars.h"               /* global variables                */
 
 #include        "calls.h"               /* generic call mechanism          */
@@ -50,9 +52,12 @@
 
 #include        "bool.h"                /* booleans                        */
 
+#include        "tls.h"                 /* thread-local storage            */
+#include        "thread.h"              /* threads                         */
+#include        "aobjects.h"            /* atomic objects                  */
+
 /****************************************************************************
 **
-
 *V  ValGVars  . . . . . . . . . . . . . . . . . .  values of global variables
 *V  PtrGVars  . . . . . . . . . . . . . pointer to values of global variables
 **
@@ -289,6 +294,10 @@ Char *          NameGVar (
     return CSTR_STRING( ELM_PLIST( NameGVars, gvar ) );
 }
 
+Obj NameGVarObj ( UInt gvar )
+{
+    return ELM_PLIST( NameGVars, gvar );
+}
 
 #define NSCHAR '@'
 
@@ -296,13 +305,13 @@ Obj CurrNamespace = 0;
 
 Obj FuncSET_NAMESPACE(Obj self, Obj str)
 {
-    CurrNamespace = str;
+    TLS(CurrNamespace) = str;
     return 0;
 }
 
 Obj FuncGET_NAMESPACE(Obj self)
 {
-    return CurrNamespace;
+    return TLS(CurrNamespace);
 }
 
 /****************************************************************************
@@ -327,7 +336,7 @@ UInt GVarName (
     Int                 len;            /* length of name                  */
 
     /* First see whether it could be namespace-local: */
-    cns = CSTR_STRING(CurrNamespace);
+    cns = CSTR_STRING(TLS(CurrNamespace));
     if (*cns) {   /* only if a namespace is set */
         len = strlen(name);
         if (name[len-1] == NSCHAR) {
@@ -1186,8 +1195,8 @@ static Int InitLibrary (
     SET_LEN_PLIST( TableGVars, SizeGVars );
 
     /* Create the current namespace: */
-    CurrNamespace = NEW_STRING(0);
-    SET_LEN_STRING(CurrNamespace,0);
+    TLS(CurrNamespace) = NEW_STRING(0);
+    SET_LEN_STRING(TLS(CurrNamespace),0);
     
     /* fix C vars                                                          */
     PostRestore( module );
@@ -1240,7 +1249,6 @@ static StructInitInfo module = {
 
 StructInitInfo * InitInfoGVars ( void )
 {
-    FillInVersion( &module );
     return &module;
 }
 
